/*
 * Copyright 2016 The OpenSSL Project Authors. All Rights Reserved.
 *
 * Licensed under the OpenSSL licenses, (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * https://www.openssl.org/source/license.html
 * or in the file LICENSE in the source distribution.
 */

/*
 * Confirm that a^b mod c agrees when calculated cleverly vs naively, for
 * random a, b and c.
 */

#include <assert.h>
#include <stdio.h>
#include <stdint.h>
#include <openssl/bn.h>
#include <openssl/err.h>

extern void LLVMFuzzerAnnounceOutput(const uint8_t *Data, size_t Size);
int LLVMFuzzerTestOneInput(const uint8_t *buf, size_t len)
{
    int success = 0;
    size_t l1 = 0, l2 = 0, l3 = 0;
    int s1 = 0, s2 = 0, s3 = 0;
    BN_CTX *ctx;
    BIGNUM *b1;
    BIGNUM *b2;
    BIGNUM *b3;
    BIGNUM *b4;

    b1 = BN_new();
    b2 = BN_new();
    b3 = BN_new();
    b4 = BN_new();
    ctx = BN_CTX_new();

    /* Divide the input into three parts, using the values of the first two
     * bytes to choose lengths, which generate b1, b2 and b3. Use three bits
     * of the third byte to choose signs for the three numbers.
     */
    if (len > 2) {
        len -= 3;
        l1 = (buf[0] * len) / 255;
        ++buf;
        l2 = (buf[0] * (len - l1)) / 255;
        ++buf;
        l3 = len - l1 - l2;

        s1 = buf[0] & 1;
        s2 = buf[0] & 2;
        s3 = buf[0] & 4;
        ++buf;
    }
    assert(BN_bin2bn(buf, l1, b1) == b1);
    BN_set_negative(b1, s1);
    assert(BN_bin2bn(buf + l1, l2, b2) == b2);
    BN_set_negative(b2, s2);
    assert(BN_bin2bn(buf + l1 + l2, l3, b3) == b3);
    BN_set_negative(b3, s3);

    uint8_t *out = NULL;
    size_t out_size = 0;

    /* mod 0 is undefined */
    if (!BN_is_zero(b3)) {
      assert(FUNC(b4, b1, b2, b3, ctx));
      out = (uint8_t*)malloc(BN_num_bytes(b4) + 1);
      out_size = BN_bn2bin(b4, out + 1) + 1;
      out[0] = BN_is_negative(b4) ? 1 : 0;
    }
    LLVMFuzzerAnnounceOutput(out, out_size);

    BN_free(b1);
    BN_free(b2);
    BN_free(b3);
    BN_free(b4);
    free(out);
    BN_CTX_free(ctx);
    ERR_clear_error();

    return 0;
}
